"""
Display management.
"""

import ctypes
from PIL import Image as PILImage
import pygame
import OpenGL
from OpenGL.GL import *
from OpenGL.GLU import *
import sys

from hitt.exceptions import UnsupportedDisplayResolution
from utils import SingletonMeta

__all__ = ('Display',)


class Display(object):
    """
    The Display class is used to set-up, test and reconfigure the display.
    """
    
    __metaclass__ = SingletonMeta
    
    _FLAGS = pygame.HWSURFACE | pygame.DOUBLEBUF | pygame.OPENGL
    
    def __init__(self):
        
        # Default values
        self._icon = None
        self._title = ''
        self._bg_color = (1.0, 1.0, 1.0, 1.0)
        self._size = None
        self._depth = None
        self._fullscreen = None
        self._surface = None
        
        self.fps = 60
        
        # Shut-down (if running) and (re)start pygame display module
        if pygame.display.get_init():
            pygame.display.quit()
        pygame.display.init()
        
        # Set default title and icon
        self.title = self._title
        self.icon = self._icon
        
        # Get a clock instance for syncing the display
        self._clock = pygame.time.Clock()
    
    # Properties
    
    def _set_title(self, value):
        self._title = value
        pygame.display.set_caption(value)
    
    def _get_title(self):
        return self._title
    
    title = property(_get_title, _set_title)
    
    def _set_icon(self, value):
        self._icon = value
        if value is None:
            pygame.display.set_icon(pygame.image.fromstring('0', (1, 1), 'P'))
        else:
            pygame.display.set_icon(pygame.image.load(value))
    
    def _get_icon(self):
        return self._icon
    
    icon = property(_get_icon, _set_icon)
    
    def _set_bg_color(self, value):
        self._bg_color = tuple([float(v) / 255.0 for v in value])
        
    def _get_bg_color(self):
        return (int(v * 255.0) for v in self._bg_color)
    
    bg_color = property(_get_bg_color, _set_bg_color)
    
    @property
    def size(self):
        return self._size
    
    @property
    def depth(self):
        return self._depth
    
    @property
    def fullscreen(self):
        return self._fullscreen
    
    @property
    def real_fps(self):
        return self._clock.get_fps()
    
    def flip_y(self, y):
        """
        Return the value of y flipped to a coordinate system based on top to 
        bottom (OpenGL default is bottom top).
        """
        return self._size[1] - y 
    
    def test(self, size, depth=32, fullscreen=False):
        """Test a display resolution, return True is supported"""
        result = pygame.display.mode_ok(size, 
                self._FLAGS | (pygame.FULLSCREEN if fullscreen else 0),
                depth)
        return result >= depth
    
    def init(self, size, depth=32, fullscreen=False):
        """Initialize the screen with specified resolution"""
        
        # Set the resolution
        try:        
            self._surface = pygame.display.set_mode(size, 
                    self._FLAGS | (pygame.FULLSCREEN if fullscreen else 0),
                    depth)
            self._size = size
            self._depth = depth
            self._fullscreen = fullscreen
        except:
            raise UnsupportedDisplayResolution((size, depth))    
    
        # Set OpenGL's rendering up for 2D
        glDisable(GL_DEPTH_TEST)
        glDisable(GL_LIGHTING)
        
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE) 
    
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluOrtho2D(0, self._size[0], 0, self._size[1])
    
    def crop(self, region=None):
        """Crop the render region"""
        if region is None:
            glDisable(GL_SCISSOR_TEST)
        else:
            glEnable(GL_SCISSOR_TEST)
            glScissor(region[0], self._size[1] - (region[1] + region[3]), region[2], region[3])
            
    def flip(self):
        """
        Flip the display buffer (back->front) and return True if the 'X' 
        (close) icon for the window has been clicked.
        """
        self._clock.tick(self.fps)
        pygame.display.flip()
        return pygame.event.get(pygame.QUIT)

    def clear(self):
        """Clear the display to the background color"""
        glClearColor(*self._bg_color) 
        glClear(GL_COLOR_BUFFER_BIT)
    
    def screenshot(self):
        """Return a screenshot of the display as a PIL image"""
        glPixelStorei(GL_PACK_ALIGNMENT, 1)
        pixel_array = glReadPixels(0, 0, 
                self._size[0], self._size[1], 
                GL_RGBA, GL_UNSIGNED_BYTE)
        return PIL.Image.new('RGBA', self._size).fromstring(pixel_array)