"""
Text rendering.

The font library supports for loading True-type fonts (.ttf) files or raw 
fonts (which are portable) created using the 'save' method of the 'Font' class
(recommended file extension .fnt).

The library renders fonts using the textured quads method. The primary goals in
developing the library where performance and flexibility, which comes at a cost 
in memory due to the large number of display lists the method generates.

Color, position, angle and scale can all be specified as well as smooth or
pixelated texture rendering.

The built-in cache option when calling the 'write' method supports for compiling
a draw array for the specified text which greatly improves performance for 
large amounts of text rendered multiple times.

True-type fonts (see utils/ttf_to_raw) are supported using PIL, which can mean
that render quality suffers. Adjusting the font size a small amount can often 
improve the quality.

Raw fonts are stored in the size they where saved at, the size cannot be
changed afterwards.
"""

import os
import struct

import Image as PILImage, ImageDraw as PILImageDraw, ImageFont as PILImageFont
from OpenGL.GL import *

from display import Display
from utils import img2tex, nearest_pow2

__all__ = ('Font',)


class Font(object):
    """
    The Font class provides text rendering support using different fonts.
    """
    
    DEFAULT_ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijsklmnopqrstuvwxyz{}()[]<>!?.,:;'\"*&%$@#\/|+=-_~`"
    
    def __init__(self, font, size=12, alphabet='', format='raw', smooth=False):
        
        self._alphabet = alphabet
        self._size = size
        self._smooth = smooth

        self._uvs = {}
        self._sizes = {}
        self._display_lists = {}
        self._cache = {}
        self._texture = None

        # Styles
        self.line_height = 0.0
        self.character_spacing = 0.0
        self.word_spacing = 0.0        

        # Load the font
        image = None
        if isinstance(font, basestring) and os.path.splitext(font)[1].lower() in ('.ttf', '.otf'):
            image = self._load_ttf(font)
        else:
            image = self._load_raw(font)
        
        # Create a texture for the font
        self._texture = img2tex(image, smooth=smooth)
        self.prepare()
    
    def __del__(self):
        # Remove any associated texture
        glDeleteTextures(self._texture)

        # Remove any associated display list
        for display_list in self._display_lists.values():
            glDeleteLists(display_list, 1)

    @property
    def alphabet(self):
        return self._alphabet

    @property
    def size(self):
        return self._size
    
    @property
    def smooth(self):
        return self._smooth
    
    def _load_raw(self, font):
        if isinstance(font, basestring):
            font = open(font, 'rb')
        
        # Alphabet
        alphabet_len = struct.unpack('i', font.read(4))[0]
        for character_index in xrange(0, alphabet_len):
            self._alphabet += struct.unpack('c', font.read(1))[0]

        # Glyphs/styles
        self._size = struct.unpack('f', font.read(4))[0]
        self.line_height = struct.unpack('f', font.read(4))[0]
        self.character_spacing = struct.unpack('f', font.read(4))[0]
        self.word_spacing = struct.unpack('f', font.read(4))[0]

        # UVs
        for character in self._alphabet:
            u1 = struct.unpack('f', font.read(4))[0]
            v1 = struct.unpack('f', font.read(4))[0]
            u2 = struct.unpack('f', font.read(4))[0]
            v2 = struct.unpack('f', font.read(4))[0]
            self._uvs[character] = (u1, v1, u2, v2)

        # Sizes
        for character in self._alphabet:
            w = struct.unpack('i', font.read(4))[0]
            h = struct.unpack('i', font.read(4))[0]
            self._sizes[character] = (w, h)

        # Image
        w = struct.unpack('i', font.read(4))[0]
        h = struct.unpack('i', font.read(4))[0]
        image = PILImage.new('L', (w, h))
        image.fromstring(font.read(w * h))
        
        font.close()
        
        return image

    def _load_ttf(self, font):
        # Load the TTF
        font = PILImageFont.truetype(font, self._size)
        
        # Create an image to render the character set to
        image = PILImage.new('L', (32, 32))
        draw = PILImageDraw.Draw(image)
        draw.setfont(font)
        
        # If not provided use the default alphabet
        if not self._alphabet:
            self._alphabet = self.DEFAULT_ALPHABET
    
        # Calculate some default glyph/style information
        alphabet_size = draw.textsize(self._alphabet)
        self._size = alphabet_size[1]
        self.line_height = self._size / 5.0 + self._size # 120% by default
        self.character_spacing = 0.0
        self.word_spacing = draw.textsize(' ')[0]
        
        # Calculate the required size of the font image
        w = nearest_pow2(alphabet_size[0])
        h = nearest_pow2(alphabet_size[1])
        
        while w >= h:
            w /= 2
            h *= 2        

        # Resize the font
        image = image.resize((w, h))
        del draw
        draw = PILImageDraw.Draw(image)
        draw.setfont(font)
        
        # Draw and map the font characters to the image
        x, y = 0, 0
        for c in self._alphabet:
            c_image = font.getmask(c).convert('L')
            c_w, c_h = c_image.size
            
            if x > w - c_image.size[0]:
                x = 0
                # Leave a single pixel gap between each row of characters
                y += self._size + 1 
            
            self._uvs[c] = (
                float(x) / float(w), 
                1.0 - float(y) / float(h), 
                float(x + c_w) / float(w), 
                1.0 - float(y + c_h) / float(h)
                )
            self._sizes[c] = (c_w, -c_h)
            
            image.paste(c_image, (x, y, x + c_w, y + c_h))
                
            x += c_w + 1
        
        return image

    def save(self, font):
        """Save the font in raw format"""
        stream = True
        if isinstance(font, basestring):
            font = open(font, 'wb')
            stream = False
            
        # Alphabet
        font.write(struct.pack('i', len(self._alphabet)))
        for c in self._alphabet:
            font.write(struct.pack('c', c))
        
        # Glyphs/styles
        font.write( struct.pack('f', self._size))
        font.write( struct.pack('f', self.line_height))
        font.write( struct.pack('f', self.character_spacing))
        font.write( struct.pack('f', self.word_spacing))
        
        # UVs
        for c in self._alphabet:
            font.write(struct.pack('f', self._uvs[c][0]))
            font.write(struct.pack('f', self._uvs[c][1]))
            font.write(struct.pack('f', self._uvs[c][2]))
            font.write(struct.pack('f', self._uvs[c][3]))

        # Sizes
        for c in self._alphabet:
            font.write(struct.pack('i', self._sizes[c][0]))
            font.write(struct.pack('i', self._sizes[c][1]))
        
        # Image
        font.write(struct.pack('i', image.size[0]))
        font.write(struct.pack('i', image.size[1]))
        font.write(image.tostring('raw', 'L', 0, 1))
    
    def prepare(self):
        """Prepare the font for rendering"""
        
        # Clear existing display lists
        self.flush_cache()
        for display_list in self._display_lists.values():
            glDeleteLists(display_list, 1)
        self._display_lists = {}
        
        # Create display lists for the alphabet
        for c in self._alphabet:
            uv0, uv1, uv2, uv3 = self._uvs[c]
            w, h = self._sizes[c]
            
            # Pre-render the mesh as a display list
            self._display_lists[c] = glGenLists(1)
            glNewList(self._display_lists[c], GL_COMPILE)
            
            glBegin(GL_TRIANGLE_FAN)

            glTexCoord2f(uv0, uv1)
            glVertex2f(0, 0)
            
            glTexCoord2f(uv2, uv1)
            glVertex2f(w, 0)
            
            glTexCoord2f(uv2, uv3)
            glVertex2f(w, h)
            
            glTexCoord2f(uv0, uv3)
            glVertex2f(0, h)
            
            glEnd()
            glEndList()
    
    def draw(self, 
            text, 
            pos, 
            angle=0.0, 
            scale=1.0, 
            color=(0, 0, 0, 255), 
            cache=False):
        """Draw the text on screen"""
        
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        
        # Position the text
        glTranslatef(pos[0], Display().flip_y(pos[1]), 1.0)
        glRotatef(angle, 0.0, 0.0, 1.0)
        glScalef(scale, scale, 1.0)
        
        # Set the color 
        glColor4ub(*color)
        
        # Select the font texture
        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, self._texture)
        
        glPushMatrix()
        
        # Check for a chached version of the text
        if text in self._cache:
            glCallList(self._cache[text])
        else:
            # Create a cache version if requested
            if cache:
                self._cache[text] = glGenLists(1)
                glNewList(self._cache[text], GL_COMPILE)
                
            x, y = 0, 0
            w = 0
            for c in text:
                # Check the character can be rendered
                if c in self._sizes:
                    glTranslatef(w, 0, 0)
                    w = self._sizes[c][0] + self.character_spacing
                    glCallList(self._display_lists[c])
                    x += w
                
                else:
                    # Check for new-line character
                    if c == '\n':
                        glTranslatef(0 - x, 0 - self.line_height, 0)
                        x = 0
                        y -= self.line_height
                        
                    else:
                        # Treat all non-renderable characters as a space ' '
                        glTranslatef(self.word_spacing, 0, 0)
                        x += self.word_spacing
                        
            if cache:
                glEndList()
                
        glPopMatrix()
        
        glDisable(GL_TEXTURE_2D)
    
    def flush_cache(self, text=None):
        """Flush the cache, optionally of the specified text"""
        if text is None:
            for text in self._cache:
                glDeleteLists(self._cache[text], 1)
            self._cache = {}
        if text in self._cache:
            glDeleteLists(self._cache[text], 1)
            del self._cache[text]
        
    def measure(self, text):
        """Measure the size of a text string"""
        lines = text.split('\n')
        width = 0

        for line in lines:
            w = 0
            
            for c in line:
                if c in self._sizes:
                    w += self._sizes[c][0] + self.character_spacing
                else:
                    w += self.word_spacing                
                    
            if w > width: 
                width = w

        height = self.line_height * len(lines)
        height -= self.line_height - self._size
        
        return (width, height)