"""
Classes and functions to help manage a game and it's various scenes (states).
"""

from hitt.utils import SingletonMeta
from scene import *

__all__ = (
    'Game',
    
    # Scenes
    'MenuScene',
    'Scene'
)


class Game(object):
    """
    The Game class is a top level singleton class responsible for managing 
    transitions between game scenes.
    """
    
    __metaclass__ = SingletonMeta
    
    def __init__(self):
        self._scenes = []
        self._running = False
    
    @property
    def running(self):
        return self._running
        
    def init(self):
        """Initialize the game"""
        self._running = True
        
    def cleanup(self):
        """Cleanup the game (e.g. free up resources, etc.)"""
        while (len(self._scenes)):
            self._scenes[-1].cleanup()
            del self._scenes[-1]

    def change_scene(self, scene):
        """Change the scene to another scene"""
        if len(self._scenes):
            self._scenes[-1].cleanup()
            del self._scenes[-1]
        self._scenes.append(scene)
        self._scenes[-1].init()

    def push_scene(self, scene):
        """Push a scene to the scene"""
        if len(self._scenes):
            self._scenes[-1].pause()
        self._scenes.append(scene)
        self._scenes[-1].init()
        
    def pop_scene(self):
        """Remove the last scene from the stack"""
        if len(self._scenes):
            self._scenes[-1].cleanup()
            del self._scenes[-1]
        
        if len(self._scenes):
            self._scenes[-1].init()
        
    def handle_events(self):
        """Handle game and scene events"""
        if len(self._scenes):
            self._scenes[-1].handle_events(self)
        
    def update(self):
        """Update the game and scene"""
        if len(self._scenes):
            self._scenes[-1].update(self)    
        
    def draw(self):
        """Draw the game and scene"""
        if len(self._scenes):
            self._scenes[-1].draw(self)
    
    def quit(self):
        """Quit the game"""
        self._running = False
        self.cleanup()