"""
A game is split into one or more scenes that each handle that part of the 
game.
"""

from hitt.collision import point_in_rect
from hitt.display import Display
from hitt.hid import HID, Keyboard
from hitt.utils import SingletonMeta

__all__ = ('MenuScene', 'Scene')


class Scene(object):
    """
    The Scene class manages a single scene (part) of the game.
    """
    
    __metaclass__ = SingletonMeta
    
    def __init__(self):
        self._paused = False
    
    @property
    def paused(self):
        return self._paused
    
    def init(self):
        """Initialize the scene"""
        self._paused = False
        
    def cleanup(self):
        """Cleanup the scene (e.g. free up resources, etc.)"""
        
    def pause(self):
        """Pause the scene"""
        self._paused = True
    
    def resume(self):
        """Resume a paused scene"""
        self._paused = False
    
    def handle_events(self, game):
        """Handle scene events"""
    
    def update(self, game):
        """Update the scene"""
        
    def draw(self, game):
        """Draw the scene"""


class MenuScene(Scene):
    """
    The MenuScene class provides a generic menu scene useful for quickly
    prototyping games.
    """
    
    def __init__(self, font, menu_items=[]):
        self.font = font
        display = Display()
        
        # Menu Items should be a list of tuples, the first element as a 
        # callback the second ('New game', new_game). The callback get's sent
        # both the game and scene instances.
        self.menu_items = menu_items
        
        self.selected_item = 0
        
    def handle_events(self, game):
        hid = HID()
        display = Display()
        
        # Keyboard
        for event in hid.keyboard.key_down_events:
            
            # Navigation
            if event.key == Keyboard.UP:
                self.selected_item -= 1
                if self.selected_item < 0:
                    self.selected_item = len(self.menu_items) - 1
            
            if event.key == Keyboard.DOWN:
                self.selected_item += 1
                if self.selected_item >= len(self.menu_items):
                    self.selected_item = 0
            
            # Selection
            if event.key == Keyboard.RETURN and len(self.menu_items):
                self.menu_items[self.selected_item][1](game, self)
            
        # Mouse
        cursor_pos = hid.mouse.pos

        h = self.font.size
        y = (display.size[1] - self.font.size * len(self.menu_items)) / 2
        for i, menu_item in enumerate(self.menu_items):
            w = self.font.measure(menu_item[0])[0]
            x = (display.size[0] - self.font.measure(menu_item[0])[0]) / 2
            
            if point_in_rect(cursor_pos, (x, y), (w, h)):
                
                # Navigation
                if len(hid.mouse.motion_events):
                    self.selected_item = i
            
                # Selection
                for event in hid.mouse.button_down_events:
                    if event.button == 1:
                        self.menu_items[self.selected_item][1](game, self)
            
            y += self.font.size
            
            # Joystick
            for joystick in hid.joysticks:
                if joystick.axis_events:
                    if joystick.axis_pos[1] < 0:
                        self.selected_item -= 1
                        if self.selected_item < 0:
                            self.selected_item = len(self.menu_items) - 1
                        
                    elif joystick.axis_pos[1] > 0:
                        self.selected_item += 1
                        if self.selected_item >= len(self.menu_items):
                            self.selected_item = 0
                
                for event in joystick.button_down_events:
                    if event.button == 1:
                        self.menu_items[self.selected_item][1](game, self)
                        
    def draw(self, game):        
        display = Display()
        
        # Clear the background
        display.bkg_color = (255, 255, 255, 255)
        display.clear()
        
        # Render the menu options
        y = (display.size[1] - self.font.size * len(self.menu_items)) / 2
        for i, menu_item in enumerate(self.menu_items):
            color = (0, 0, 0, 255)
            if i == self.selected_item:
                color = (0, 0, 255, 255)
            x = (display.size[0] - self.font.measure(menu_item[0])[0]) / 2
            self.font.draw(menu_item[0], 
                (x, y + i * self.font.size), 
                color=color, 
                cache=True)