"""
Human interface devices.
"""

import pygame
import pygame.locals

from utils import SingletonMeta

__all__ = ('HID', 'Keyboard', 'Mouse', 'Joystick')


class HID(object):
    """
    The 'HID' class is a singleton class designed to simplify the management 
    of the keyboard, mouse and joysticks.
    """
    
    __metaclass__ = SingletonMeta
    
    def __init__(self):
        
        # Initialize joystick support in PyGame
        if pygame.joystick.get_init():
            pygame.joystick.quit()
        pygame.joystick.init()
        
        self._events = []
        self._devices = {}
        
        self.query()
    
    @property
    def keyboard(self):
        return self._devices.get('keyboard', None)
    
    @property
    def mouse(self):
        return self._devices.get('mouse', None)
        
    @property
    def joysticks(self):
        return self._devices.get('joysticks', None)

    def query(self):
        """Query for HIDs"""
        
        # Only a single keyboard and mouse device is supported and they're 
        # always available
        self._devices = {
            'keyboard': Keyboard(),
            'mouse': Mouse(),
            }
        
        # Poll for joysticks    
        joysticks = []
        joystick_count = pygame.joystick.get_count()
        for joystick_id in xrange(0, joystick_count):
            joystick = Joystick(joystick_id)
            joysticks.append(joystick)
        
        self._devices['joysticks'] = tuple(joysticks)
    
    def update(self):
        """Should be called once per frame before the display flip"""
        
        # Update the devices
        self.keyboard.update()
        self.mouse.update()
        
        # Joystick events have to be handled differently as events cannot be
        # polled per joystick.
        button_down_events = pygame.event.get(pygame.JOYBUTTONDOWN)
        button_up_events = pygame.event.get(pygame.JOYBUTTONUP)
        axis_events = pygame.event.get(pygame.JOYAXISMOTION)
        hat_events = pygame.event.get(pygame.JOYHATMOTION)
        trackball_events = pygame.event.get(pygame.JOYBALLMOTION)
        events = (
            button_down_events,
            button_up_events,
            axis_events,
            hat_events,
            trackball_events
            )
        
        [joystick.update(*events) for joystick in self.joysticks]
        
        pygame.event.pump()
        

# Note: Keybaord, Mouse and Joystick class instances should be accessed 
# through the HID class and not directly.

    
class Keyboard(object):
    """
    The Keyboard class supports querying of keyboard events.
    """
    
    def __init__(self):
        self.key_down_events = None
        self.key_up_events = None
    
    @property
    def has_focus(self):
        return pygame.key.get_focused()
    
    # repeat property
    def _get_repeat(self):
        return pygame.key.get_repeat()
        
    def _set_repeat(self, value):
        if value:
            pygame.key.set_repeat(*value)
        else:
            pygame.key.set_repeat()
    
    repeat = property(_get_repeat, _set_repeat)

    # modifiers property
    def _get_modifiers(self):
        return pygame.key.get_mods()
        
    def _set_modifiers(self, value):
        pygame.key.set_mods(value)
        
    modifiers = property(_get_modifiers, _set_modifiers)
    
    def key_down(self, key):
        """Return True it the specified key is pressed down"""
        return pygame.key.get_pressed()[key]
    
    def update(self):
        """Update the keyboard with the current frame of events"""
        self.key_down_events = pygame.event.get(pygame.KEYDOWN)
        self.key_up_events = pygame.event.get(pygame.KEYUP)
        
    
# Update the Keyboard class to hold a copy of all the pygame key variables
for attr in dir(pygame.locals):
    if attr.startswith('K_'):
        setattr(Keyboard, attr.lstrip('K_'), getattr(pygame.locals, attr))


class Mouse(object):
    """
    The Mouse class supports querying of mouse events.
    """
    
    def __init__(self):
        self._visible = True
        self.button_down_events = None
        self.button_up_events = None
        self.motion_events = None
        self.rel = (0.0, 0.0)

    @property
    def has_focus(self):
        return pygame.mouse.get_focused()

    @property
    def pos(self):
        return pygame.mouse.get_pos()
    
    # visible property
    def _get_visible(self):
        return self._visible
    
    def _set_visible(self, value):
        self._visible = visible
        pygame.mouse.set_visible(value)
    
    def button_down(self, button):
        """Return True is the specified button is pressed down"""
        return pygame.mouse.get_pressed()[button]
        
    def update(self):
        """Update the mouse with the current frame of events"""
        self.button_down_events = pygame.event.get(pygame.MOUSEBUTTONDOWN)
        self.button_up_events = pygame.event.get(pygame.MOUSEBUTTONUP)
        self.motion_events = pygame.event.get(pygame.MOUSEMOTION)
        self.rel = pygame.mouse.get_rel()

        
class Joystick(object):
    """
    The Joystick class supports querying of joystick events.
    """
    
    def __init__(self, id):
        # Initialize the joystick
        self._joystick = pygame.joystick.Joystick(id)
        self._joystick.init()
        
        self.id = id
        self.name = self._joystick.get_name()
        self.axis_count = self._joystick.get_numaxes()
        self.button_count = self._joystick.get_numbuttons()
        self.hat_count = self._joystick.get_numhats()
        self.trackball_count = self._joystick.get_numballs()
        
        # Events
        self.button_down_events = None
        self.button_up_events = None
        self.axis_events = None
        self.hat_events = None
        self.trackball_events = None

    def button_is_down(self, button):
        """Return True if the specified button is pressed down"""
        return bool(self._joystick.get_button(button))
    
    def axis_pos(self, axis, tol=0.05):
        """Return the position of the specified axis"""
        pos = self._joystick.get_axis(axis)
        if (abs(pos[0]) < tol):
            return 0
        return pos
    
    def hat_pos(self, hat, tol=0.05):
        """Return the position of the specified hat (x, y)"""
        xy = self._joystick.get_hat(hat)
        xy = [xy[0], xy[1]]
        
        if (abs(xy[0]) < tol):
            xy[0] = 0        
        if (abs(xy[1]) < tol):
            xy[1] = 0

        return tuple(xy)

    def get_trackball(self, trackball):
        """Return the position of the specified trackball (x, y)"""
        xy = self._joystick.get_ball(trackball)
        xy = [xy[0], xy[1]]
        
        if (abs(xy[0]) < tol):
            xy[0] = 0        
        if (abs(xy[1]) < tol):
            xy[1] = 0

        return tuple(xy)
    
    def update(self, 
        button_up_events, 
        button_down_events,
        axis_events,
        hat_events,
        trackball_events):
        
        """Update the joystick with the current frame of events"""
        self.button_down_events = [e for e in button_down_events if e.joy == self.id] 
        self.button_up_events = [e for e in button_up_events if e.joy == self.id]
        self.axis_events = [e for e in axis_events if e.joy == self.id]
        self.hat_events = [e for e in hat_events if e.joy == self.id]
        self.trackball_events = [e for e in trackball_events if e.joy == self.id]
    