"""
Image rendering.
"""

from OpenGL.GL import *
from PIL import Image as PILImage

from display import Display
from utils import img2tex, nearest_pow2, split_pow2, MixinMeta
    
__all__ = ('Image', 'Sprite')


class ImageMixin(object):
    """
    The ImageMixin class provides common image functionality.
    """
    
    # handle property
    def _get_handle(self):
        return self._handle
    
    def _set_handle(self, value):
        self._handle = value
        self.prepare()
    
    handle = property(_get_handle, _set_handle)
    
    @property
    def size(self):
        return self._size
    
    @property
    def mode(self):
        return self._mode
        
    @property
    def smooth(self):
        return self._smooth
    
    def center_handle(self):
        """Center the sprites handle""" 
        self.handle = (-self._size[0] / 2, -self._size[1] / 2)

    def draw(self):
        """Draw the image on screen"""
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        # Transform the image
        glTranslatef(self.x, Display().flip_y(self.y), 1.0)
        glRotatef(self.angle, 0.0, 0.0, -1.0)
        glScalef(self.scale, self.scale, 1.0)

        # Render the image
        glColor4ub(*self.color)
        glCallList(self._display_list)    
    

class Sprite(object):
    """
    Sprites more restrictive (pow2 sizes only) but fast.
    """
    
    __metaclass__ = MixinMeta
    __mixins__ = (ImageMixin,)
    
    def __init__(self, image, smooth=False):
        # Defaults
        self.x = 0.0
        self.y = 0.0
        self.angle = 0.0     
        self.scale = 1.0
        self._handle = (0.0, 0.0)
        self.color = (255, 255, 255, 255)

        # Convert to a PIL image if not already
        if isinstance(image, basestring):
            image = PILImage.open(image)
    
        # Remember for size, mode and smooth
        self._size = image.size
        self._mode = image.mode
        self._smooth = smooth

        # Sprite will be prepare into a display list
        self._display_list = None

        self._texture = img2tex(image, smooth)

        self.prepare()
    
    def __del__(self):
        # Remove any associated texture
        if self._texture is not None:
            glDeleteTextures([self._texture])

        # Remove any associated display list
        if self._display_list is not None:
            glDeleteLists(self._display_list, 1)
    
    def prepare(self):
        """Prepare the sprite for drawing"""
        
        # Create a display list to store the sprite drawing operations in
        if self._display_list is not None:
            glDeleteLists(self._display_list, 1)
        self._display_list = glGenLists(1)
        glNewList(self._display_list, GL_COMPILE)
        
        # Pre-render the sprite
        glPolygonMode(GL_FRONT, GL_FILL)      
        glEnable(GL_TEXTURE_2D)  
        
        # Assign the texture
        glBindTexture(GL_TEXTURE_2D, self._texture)
       
        # Draw the quad
        glBegin(GL_TRIANGLE_FAN)
        
        glTexCoord2i(0, 1)
        glVertex2f(self._handle[0], -self._handle[1])
        
        glTexCoord2i(1, 1)
        glVertex2f(self._size[0] + self._handle[0], -self._handle[1])
        
        glTexCoord2i(1, 0)
        glVertex2f(self._size[0] + self._handle[0], -self._size[1] - self._handle[1])
        
        glTexCoord2i(0, 0)
        glVertex2f(self._handle[0], -self._size[1] - self._handle[1])
        
        glEnd()
        glDisable(GL_TEXTURE_2D)
        glEndList()
        

class Image(object):
    """
    Images support images of any size but are slow (relative to sprites).
    """
    
    __metaclass__ = MixinMeta
    __mixins__ = (ImageMixin,)

    def __init__(self, image, smooth=False):
        # Defaults
        self.x = 0.0
        self.y = 0.0
        self.angle = 0.0     
        self.scale = 1.0
        self._handle = (0.0, 0.0)
        self.color = (255, 255, 255, 255)

        # Convert to a PIL image if not already
        if isinstance(image, basestring):
            image = PILImage.open(image)

        # Remember for size, mode and smooth
        self._size = image.size
        self._mode = image.mode
        self._smooth = smooth

        # Sprite will be prepare into a display list
        self._display_list = None

        # To support for variable sizes the system cuts the image into
        # multiple pow2 quads.
        self._textures = []
        self._quads = []

        rows = split_pow2(self.size[1])
        cols = split_pow2(self.size[0])
        
        row_pos = 0
        for row_index in xrange(0, len(rows)):
            
            self._textures.append([])
            self._quads.append([])
            row = rows[row_index]
            col_pos = 0
            
            for col_index in xrange(0, len(cols)):
                col = cols[col_index]
                
                # Store the coords
                self._quads[row_index].append((col_pos, row_pos, col, row))
                
                # Create a texture for the quad
                quad_image = image.crop((col_pos, row_pos, 
                    col_pos + col, row_pos + row))
                
                texture = img2tex(quad_image, smooth)
                self._textures[row_index].append(texture)
                
                col_pos += col
            row_pos += row
                
        self.prepare()
    
    def __del__(self):
       # Remove any associated textures
       for row in self._textures:
           glDeleteTextures(row)

       # Remove any associated display list
       if self._display_list is not None:
           glDeleteLists(self._display_list, 1)

    def prepare(self):
        """Prepare the image for drawing"""
                   
        # Create a display list to store the image drawing operations in
        if self._display_list is not None:
            glDeleteLists(self._display_list, 1)
        self._display_list = glGenLists(1)
        glNewList(self._display_list, GL_COMPILE)

        # Pre-render the image
        glPolygonMode(GL_FRONT, GL_FILL)      
        glEnable(GL_TEXTURE_2D)

        for row_index in range(0, len(self._quads)):
            for col_index in range(0, len(self._quads[row_index])):
                quad = self._quads[row_index][col_index]
                texture = self._textures[row_index][col_index]

                # Assign the texture
                glBindTexture(GL_TEXTURE_2D, texture)

                # Draw the quad
                glBegin(GL_TRIANGLE_FAN)

                glTexCoord2i(0, 1)
                glVertex2f(quad[0] + self._handle[0], 
                    -(quad[1] + self._handle[1]))

                glTexCoord2i(1, 1)
                glVertex2f(quad[0] + quad[2] + self._handle[0], 
                    -(quad[1] + self._handle[1]))

                glTexCoord2i(1, 0)
                glVertex2f(quad[0] + quad[2] + self._handle[0],
                    -(quad[1] + quad[3] + self._handle[1]))

                glTexCoord2i(0, 0)
                glVertex2f(quad[0] + self._handle[0],
                    -(quad[1] + quad[3] + self._handle[1]))

                glEnd()  

        glDisable(GL_TEXTURE_2D)
        glEndList()
        