"""
Tile management classes.
"""

from numpy import *
from OpenGL.GL import *
from PIL import Image as PILImage
import struct
import time

from hitt.display import *
from hitt.resource import *
from hitt.shape import Rect
from hitt.utils import img2tex
from hitt.utils import bin2img, bin2str, img2bin, str2bin


__all__ = (
    # Classes
    'Tile',
    'TileLayer',
    'TileMap',
    'TileSet'
    )


class Tile(object):
    """
    A base tile class for tiles in a TileLayer.
    """

    def __init__(self, tile_set_index=0, tile_swatch=None, solid=False):
        self.tile_set_index = tile_set_index
        
        if tile_swatch is None:
            tile_swatch = [0, 0]
        self.tile_swatch = tile_swatch
        self.solid = solid
        
        # Draw is an internal flag used to determine if the tile needs to be
        # drawn when being rendered in the layer. Grouping and occlusion 
        # determine this flags value.
        self._draw = True
    
    def save(self):
        """Return a data string that can be used to save the tile"""
        return struct.pack('bBB?', 
            self.tile_set_index, 
            self.tile_swatch[0], 
            self.tile_swatch[1], 
            self.solid)
    
    @classmethod
    def load(cls, data):
        """Return a tile loaded from a data string"""
        tile_set_index, swatch_x, swatch_y, solid = struct.unpack('bBB?', data)
        return cls(tile_set_index, (swatch_x, swatch_y), solid)


class TileLayer(object):
    """
    A layer of tiles within a TileMap.
    """
    
    TILE_CLS = Tile
    
    DRAW_CLIPPED = 1
    DRAW_CHUNKED = 2
    
    def __init__(self, size, draw_method=DRAW_CLIPPED):
        
        self._size = size
        
        # Create an array in which to store the tile information
        self.tiles = zeros(size[0] * size[1], dtype=object).reshape(*size)
        
        # Fill the array with None values
        for i in xrange(0, size[0] * size[1]):
            y, x = divmod(i, size[0])
            self.tiles[x][y] = None
            
        # There are mulitple ways of rendering the tile layer, the technique 
        # used is determined by the draw_method flag. The attributes below 
        # are specific to different draw methods.
        
        # Note: if the draw method for a layer is modified then the map should
        # be prepared again.
        self.draw_method = draw_method
        
        # Chunked rendering
        self.chunk_size = 10
        self._chunks = None
        self._chunk_map = None
    
    def __del__(self):    
        self._cleanup_chunked()

    @property
    def size(self):
        return self._size
    
    # Public methods
    
    def fill(self, tiles, region=None):
        """Fill the layer (or a region of the layer) with tiles"""
        
        if region is None:
            region = (0, 0, self._size[0], self._size[1])
        
        width = (region[2] - region[0])
        tile_count = width * (region[3] - region[1])
        
        for i in xrange(0, tile_count):
            y, x = divmod(i, width)
            self.tiles[x + region[0]][y + region[1]] = tiles[i]
            
    def clear(self, region=None):
        """Clear the layer (or a region of the layer)"""
        
        if region is None:
            region = (0, 0, self._size[0], self._size[1])
        
        width = (region[2] - region[0])
        tile_count = width * (region[3] - region[1])
        
        for i in xrange(0, tile_count):
            y, x = divmod(i, width)
            self.tiles[x + region[0]][y + region[1]] = None
        
    def default_fill(self, tile_class, region=None, *args, **kw):
        """
        Fill the layer (or a region of the layer) with tiles of the specified 
        class.
        """
        
        if region is None:
            region = (0, 0, self._size[0], self._size[1])
        
        width = (region[2] - region[0])
        tile_count = width * (region[3] - region[1])
        
        for i in xrange(0, tile_count):
            y, x = divmod(i, width)
            self.tiles[x + region[0]][y + region[1]] = tile_class(*args, **kw)       

    def save(self):
        """Return a data string that can be used to save the tile layer"""
        data = struct.pack('HH', *self._size)
        for i in xrange(0, self._size[0] * self._size[1]):
            y, x = divmod(i, self._size[0])
            tile = self.tiles[x][y]
            if tile is None:
                data += struct.pack('I', 0)
            else:
                tile_data = self.tiles[x][y].save()
                data += struct.pack('I', len(tile_data))
                data += tile_data
        return data

    @classmethod
    def load(cls, data):
        """Return a tile layer loaded from a data string"""
        size = struct.unpack('HH', data[:struct.calcsize('HH')])
        data = data[struct.calcsize('HH'):]

        tiles = []
        for i in xrange(0, size[0] * size[1]):
            tile_length = struct.unpack('I', data[:struct.calcsize('I')])[0]
            data = data[struct.calcsize('I'):]
            if tile_length == 0:
                tiles.append(None)
            else:
                tiles.append(cls.TILE_CLS.load(data[:tile_length]))
                data = data[tile_length:]

        tile_layer = cls(size)
        tile_layer.fill(tiles)

        return tile_layer
        
    # Private methods

    def _prepare(self, tile_map):
        """Prepare the tile layer for drawing"""
        
        # Clean up previous prepared data
        self._cleanup_chunked()
        
        if self.draw_method == TileLayer.DRAW_CHUNKED:
            self._prepare_chunked(tile_map)

    def _draw(self, tile_map, region):
        """Draw the tile layer"""
        
        # What method do we use to draw the tile layer?
        if self.draw_method == TileLayer.DRAW_CLIPPED:
            self._draw_clipped(tile_map, region)
        elif self.draw_method == TileLayer.DRAW_CHUNKED:
            self._draw_chunked(tile_map, region)
    
    # Drawing methods
    
    def _draw_clipped(self, tile_map, region):
        """Draw a clipped version of the map"""
        size = self._size
        tile_size = tile_map._tile_size
        
        x = tile_map.x + tile_size
        y = tile_map.y + tile_size

        x_count = divmod(region[2] + tile_size * 3, tile_size)[0]
        y_count = divmod(region[3] + tile_size * 3, tile_size)[0]

        x_tile, x_offset = divmod(-x, tile_size)
        y_tile, y_offset = divmod(-y, tile_size)

        x_region = region[0]
        y_region = region[1]

        for i in xrange(0, x_count * y_count):
            y_inner, x_inner = divmod(i, x_count)

            x_index = x_tile + x_inner
            y_index = y_tile + y_inner

            # Don't attempt to render anything outside of the later
            if x_index < 0 or x_index > size[0] - 1 or y_index < 0 or y_index > size[1] - 1:
                continue

            tile = self.tiles[x_index][y_index]
            if tile is None:
                continue
    
            x_draw = tile_map.x + x_region + x_index * tile_size
            y_draw = tile_map.y + y_region + y_index * tile_size

            if tile._draw:
                swatch = tile_map.tile_sets[tile.tile_set_index]
                swatch.draw(tile.tile_swatch, (x_draw, y_draw))
            
            # Editor overlays
            if tile_map.overlay_solids and tile.solid:
                tile_map._overlay.fill_color = tile_map.solid_overlay_color
                tile_map._overlay.x = x_draw
                tile_map._overlay.y = y_draw
                tile_map._overlay.draw()

    def _cleanup_chunked(self):
        """Clean up any data used to prepare the layer for chunked rendering"""
        
        # Delete any existing display lists
        if self._chunks is not None:
            for chunk in self._chunks.flat:
                glDeleteLists(chunk[1], 1)
        
        self._chunks = None
        self._chunk_map = None
        
    def _prepare_chunked(self, tile_map):
        """Prepare the layer for drawing in chunks"""
    
        size = self.size
        tile_size = tile_map.tile_size
        chunk_size = self.chunk_size
        tile_count = chunk_size * chunk_size
        
        # Divide the grid into chunks
    
        chunk_grid_size = [
            divmod(size[0], chunk_size)[0], 
            divmod(size[1], chunk_size)[0]
            ]
        
        if size[0] % chunk_size != 0:
            chunk_grid_size[0] += 1

        if size[1] % chunk_size != 0:
            chunk_grid_size[1] += 1    
    
        self._chunks = zeros(
            chunk_grid_size[0] * chunk_grid_size[1], 
            dtype=object
            ).reshape(*chunk_grid_size)
        
        self._chunk_map = {}
        
        for y_chunk in range(0, chunk_grid_size[0]):
            for x_chunk in range(0, chunk_grid_size[1]):
                
                x_origin = x_chunk * chunk_size
                y_origin = y_chunk * chunk_size 
                
                # Create a display list of drawn tiles for each chunk
                display_list = glGenLists(1)
                glNewList(display_list, GL_COMPILE)
                
                # Store the display list
                self._chunks[x_chunk][y_chunk] = ((x_origin, y_origin), display_list)
                chunk_index = len(self._chunks) - 1
                
                for y_inner in range(0, chunk_size):
                    for x_inner in range(0, chunk_size):
                        
                        x_index = x_origin + x_inner
                        y_index = y_origin + y_inner
                        
                        # Remember what chunk this tile references
                        self._chunk_map[(x_index, y_index)] = (x_chunk, y_chunk)
                                                    
                        draw = True
                        
                        # Is the index within the boundary of the map?
                        if x_index > size[0] - 1 or y_index > size[1] - 1:
                            draw = False 
                        
                        # Is there a tile at that index?
                        tile = None
                        if draw:
                            tile = self.tiles[x_index][y_index]
                        
                        if tile is None:
                            draw = False
                        
                        # Should we draw the tile?
                        if draw and tile._draw:
                            swatch = tile.tile_swatch
                            tile_set = tile_map.tile_sets[tile.tile_set_index]
                            display_lists = tile_set._display_lists
                            glCallList(display_lists[swatch[0]][swatch[1]])
                            
                        # Move one tile along
                        glTranslatef(tile_size, 0, 0)
                    
                    # Move one tile down and to the start of the chunk across
                    glTranslatef(-tile_size * chunk_size, -tile_size, 0)
                
                # Move down a row
                glTranslatef(0, tile_size * chunk_size, 0)

                glEndList()
                
    def _draw_chunked(self, tile_map, region):
        """
        Draw the tiles in chunks.
        
        This method breaks the map down into chunks of tiles, each chunk is 
        rendered as a display list, reducing the render calls.
        
        The method is best suited for rendering static layers that where the 
        z-index order of the tiles is of no importance.
        """
        
        size = self._size
        tile_size = tile_map._tile_size
        
        x = tile_map.x + tile_size
        y = tile_map.y + tile_size

        x_count = divmod(region[2] + tile_size * 3, tile_size)[0]
        y_count = divmod(region[3] + tile_size * 3, tile_size)[0]

        x_tile, x_offset = divmod(-x, tile_size)
        y_tile, y_offset = divmod(-y, tile_size)

        chunked = []
        
        glColor4ub(255, 255, 255, 255)
        
        x_first = min(self._size[0] - 1, max(0, x_tile))
        y_first = min(self._size[1] - 1, max(0, y_tile))
        x_last = min(self._size[0] - 1, max(0, x_tile + x_count))
        y_last = min(self._size[1] - 1, max(0, y_tile + y_count))
        
        first_chunk = self._chunk_map[(x_first, y_first)]
        last_chunk = self._chunk_map[(x_last, y_last)]
        
        x_count = last_chunk[0] - first_chunk[0] + 1
        y_count = last_chunk[1] - first_chunk[1] + 1
        
        for i in xrange(0, x_count * y_count):
            y_inner, x_inner = divmod(i, x_count)
            
            chunk = self._chunks[first_chunk[0] + x_inner][first_chunk[1] + y_inner]
            
            glMatrixMode(GL_MODELVIEW)
            glLoadIdentity()
            glTranslatef(
                tile_map.x + chunk[0][0] * tile_size, 
                Display().flip_y(tile_map.y + chunk[0][1] * tile_size), 
                1.0)
            
            # Draw the chunk
            display_list = chunk[1]
            glCallList(display_list)
            
            
class TileSet(object):
    """
    A set of tiles.
    """
    
    def __init__(self, size, tile_size, image):
        self._size = size
        self._swatch_count = size[0] * size[1]
        self._tile_size = tile_size
        
        # Convert to a PIL image if not already
        if isinstance(image, basestring):
            image = PILImage.open(image)

        # Remember the image for the purpose of saving
        self._image = image

        # Generate a texture for the tileset
        self._texture = img2tex(image, False)
        self._overlay = None
        
        self._display_lists = None
        self.prepare()

    def __del__(self):
        # Remove any associated texture
        if self._texture is not None:
            glDeleteTextures([self._texture])

        # Remove any associated display lists
        if self._display_lists is not None:
            for i in xrange(0, self._swatch_count):
                y, x = divmod(i, self._size[0])
                glDeleteLists(self._display_lists[x][y], 1)

    @property
    def size(self):
        return self._size

    @property
    def swatch_count(self):
        return self._swatch_count

    @property
    def tile_size(self):
        return self._tile_size

    def prepare(self):
        """Prepare the tileset for drawing"""
        
        swatch_count = self._swatch_count
        tile_size = self._tile_size

        # Delete any existing display lists
        if self._display_lists is not None:
            for i in xrange(0, swatch_count):
                y, x = divmod(i, self._size[0])
                glDeleteLists(self._display_lists[x][y], 1)

        # Create an array in which to store the display lists
        self._display_lists = arange(swatch_count).reshape(*self._size)

        texture_size = self._image.size
        texture_width = float(tile_size) / float(texture_size[0])
        texture_height = float(tile_size) / float(texture_size[1])

        # Pre-render the tiles
        for i in xrange(0, swatch_count):
            y, x = divmod(i, self._size[0])

            # Create a display list to store the image drawing operations in
            self._display_lists[x][y] = glGenLists(1)
            glNewList(self._display_lists[x][y], GL_COMPILE)

            # Pre-render the tile
            glPolygonMode(GL_FRONT, GL_FILL)      
            glEnable(GL_TEXTURE_2D)

            # Assign the texture
            glBindTexture(GL_TEXTURE_2D, self._texture)

            # Draw the quad
            glBegin(GL_TRIANGLE_FAN)

            glTexCoord2f(x * texture_width, -(y * texture_height))
            glVertex2f(0, 0)

            glTexCoord2f(x * texture_width + texture_width, -(y * texture_height))
            glVertex2f(tile_size, 0)

            glTexCoord2f(x * texture_width + texture_width, -(y * texture_height + texture_height))
            glVertex2f(tile_size, -self._tile_size)

            glTexCoord2f(x * texture_width, -(y * texture_height + texture_height))
            glVertex2f(0, -tile_size)

            glEnd()  

            glDisable(GL_TEXTURE_2D)
            glEndList()

    def draw(self, tile, pos):
        """Draw the image on screen"""
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        # Transform the image
        glTranslatef(pos[0], Display().flip_y(pos[1]), 1.0)

        # Render the image
        glColor4ub(255, 255, 255, 255)
        glCallList(self._display_lists[tile[0]][tile[1]])
        
    def save(self):
        """Return a data string that can be used to save the tile set"""
        data = struct.pack('HH', *self._size)
        data += struct.pack('H', self._tile_size)
        data += img2bin(self._image)
        return data

    @classmethod
    def load(cls, data):
        """Return a tile set loaded from a data string"""
        size = struct.unpack('HH', data[:struct.calcsize('HH')])
        data = data[struct.calcsize('HH'):]
        tile_size = struct.unpack('H', data[:struct.calcsize('H')])[0]
        data = data[struct.calcsize('H'):]
        image = bin2img(data)
        return cls(size, tile_size, image)


class TileMap(object):
    """
    A class for storing a 2D map of tiles.
    """
    
    TILE_SET_CLS = TileSet
    TILE_LAYER_CLS = TileLayer
    
    def __init__(self, size, tile_size):

        self._size = size
        self._tile_count = size[0] * size[1]

        self._tile_size = tile_size

        self.x = 0
        self.y = 0

        self.tile_sets = []
        self.tile_layers = []
        
        # Overlay flags - Useful for editors
        self.overlay_solids = False
        self.solid_overlay_color = (255, 0, 0, 64)

        # Create a shape for highlighting tiles
        self._overlay = Rect((tile_size, tile_size))
        self._overlay.outline_color = None
        
        self.prepare()

    @property
    def size(self):
        return self._size

    @property
    def tile_size(self):
        return self._tile_size

    def index(self, pos, region=None):
        """Return the tile index at the specified screen position"""

        if region is None:
            region = (0, 0)

        x = pos[0] - region[0] - self.x
        y = pos[1] - region[1] - self.y

        x_tile = divmod(x, self._tile_size)[0]
        y_tile = divmod(y, self._tile_size)[0]

        # Return empty result if pos is outside of map
        if x_tile < 0 or x_tile > self._size[0] - 1 or y_tile < 0 or y_tile > self._size[1] - 1:
            return None

        return (x_tile, y_tile)

    def position(self, index, region=None):
        """Return the screen position of a tile for the specified index"""
        x = self.x + self._tile_size * index[0] 
        y = self.y + self._tile_size * index[1]

        return [x, y]

    def prepare(self):
        """Prepare the map for drawing"""
        for tile_layer in self.tile_layers:
            tile_layer._prepare(self)
        
    def draw(self, region=None):
        """Draw the map"""

        if region is None:
            region = (0, 0, Display().size[0], Display().size[1])

        for tile_layer in self.tile_layers:
            tile_layer._draw(self, region)

    def save(self):
        """Return a data string that can be used to save the tile map"""
        # Size
        data = struct.pack('HH', *self._size)

        # Tile size
        data += struct.pack('H', self._tile_size)

        # Position
        data += struct.pack('ii', self.x, self.y)

        # Tile sets
        data += struct.pack('B', len(self.tile_sets))
        for tile_set in self.tile_sets:
            tile_set_data = tile_set.save()
            data += struct.pack('Q', len(tile_set_data))
            data += tile_set_data

        # Tile layers
        data += struct.pack('B', len(self.tile_layers))
        for tile_layer in self.tile_layers:
            tile_layer_data = tile_layer.save()
            data += struct.pack('Q', len(tile_layer_data))
            data += tile_layer_data

        return data

    @classmethod
    def load(cls, data):
        """Return a tile map loaded from a data string"""
        # Size
        size = struct.unpack('HH', data[:struct.calcsize('HH')])
        data = data[struct.calcsize('HH'):]

        # Tile size
        tile_size = struct.unpack('H', data[:struct.calcsize('H')])[0]
        data = data[struct.calcsize('H'):]

        # Create the tile map
        tile_map = TileMap(size, tile_size)

        # Position
        x, y = struct.unpack('ii', data[:struct.calcsize('ii')])
        data = data[struct.calcsize('ii'):]
        tile_map.x = x
        tile_map.y = y

        # Tile sets
        tile_set_count = struct.unpack('B', data[:struct.calcsize('B')])[0]
        data = data[struct.calcsize('B'):]
        for i in xrange(0, tile_set_count):
            tile_set_length = struct.unpack('Q', data[:struct.calcsize('Q')])[0]
            data = data[struct.calcsize('Q'):]
            tile_map.tile_sets.append(cls.TILE_SET_CLS.load(data[:tile_set_length]))
            data = data[tile_set_length:]

        # Tile layers
        tile_layer_count = struct.unpack('B', data[:struct.calcsize('B')])[0]
        data = data[struct.calcsize('B'):]
        for i in xrange(0, tile_layer_count):
            tile_layer_length = struct.unpack('Q', data[:struct.calcsize('Q')])[0]
            data = data[struct.calcsize('Q'):]
            tile_map.tile_layers.append(cls.TILE_LAYER_CLS.load(data[:tile_layer_length]))
            data = data[tile_layer_length:]        

        return tile_map