"""
Resource management.
"""

from font import *
from image import *
from utils import SingletonMeta


__all__ = ('Resource', 'ResourceMgr')


class ResourceMgr(object):
    """
    The ResourceMgr class manages access to resources.
    """
    
    __metaclass__ = SingletonMeta
    
    def __init__(self):
        self._types = {}
        self._resources = {}
    
    @property
    def types(self):
        return self._types
    
    @property
    def resources(self):
        return self._resources
    
    def _add_resource(self, resource):
        """Add a resource to the manager"""
        self._resources[resource.id] = resource
        
    def _remove_resource(self, resource):
        """Remove a resource from the manager"""
        del self._resources[resource.id]
    
    def add_type(self, name, cls):
        """Add a resource type to the manager"""
        if name in self._types and cls is not self._types[name]:
            raise ResourceTypeDefined((name, cls))
        self._types[name] = cls
    
    def remove_type(self, name):
        """Remove a resource type from the manager"""
        if name in self._types:
            # Remove all resources of that type
            rids = self._resources.keys()
            for rid in rids:
                resource = self._resources[rid]
                if resource.type == name:
                    del self._resources[rid]
            
            # Remove the type
            del self._types[name]
    
    def remove_all_resources(self):
        """Remove all resources from the manager"""
        rids = self._resources.keys()
        for rid in rids:
            del self._resources[rid]

# Add default resource types
ResourceMgr().add_type('Font', Font)
ResourceMgr().add_type('Image', Image)
ResourceMgr().add_type('Sprite', Sprite)


class Resource(object):
    """
    The Resource class manages to loading and release of resources.
    """
    
    IMMEDIATE = 1
    DEFERRED = 2
    
    def __init__(self, rid, rtype, rload=IMMEDIATE, *args, **kw):
        
        # Is the ID unique?
        if rid in ResourceMgr().resources:
            raise ResourceIDNotUnique(rid)
        
        # Is the type recognised ?
        if not rtype in ResourceMgr().types:
            raise UnrecognisedResourceType(rtype)
        
        self._id = rid
        self._type = rtype
        self._load = rload
        self._args = args
        self._kw = kw
        self._loaded = False
        self._instance = None
        
        if rload == Resource.IMMEDIATE:
            self.get()
        
        ResourceMgr()._add_resource(self)
    
    @property
    def id(self):
        return self._id
    
    @property
    def type(self):
        return self._type
    
    @property
    def load(self):
        return self._load
    
    @property
    def loaded(self):
        return self._loaded
    
    @property
    def args(self):
        return self._args
    
    @property
    def kw(self):
        return self._kw
    
    def get(self):
        """Get a resource (loaded if necessary)"""
        if not self._loaded:
            cls = ResourceMgr().types[self._type]
            self._instance = cls(*self.args, **self.kw)
            self._loaded = True
        return self._instance
        
    def release(self):
        """Release the resource"""
        del self._instance
        self._loaded = False
    
    def __del__(self):
        del self._instance
        ResourceMgr()._remove_resource(self)

