"""
Shapes.
"""

# Make shapes support there own position and colours, makes them easier to match to images and sprites and makes the interface more uniform

from math import atan2, cos, radians, sin, sqrt, pi
from OpenGL.GL import *
import struct

from display import Display

__all__ = (
    'Ellipse',
    'Line',
    'Point',
    'Polygon',
    'Rect'
    )


class Shape(object):
    """
    The Shape class is a base class for all shapes.
    """
    
    def __init__(self, points, smooth=False):
        # Defaults
        self.x = 0.0
        self.y = 0.0
        self.angle = 0.0
        self.scale = 1.0
        self._handle = (0.0, 0.0)
        self.outline_color = (0, 0, 0, 255)
        self.fill_color = None
        
        self._smooth = smooth
        
        # The point(s) that make up the shape
        self._points = points
        
        # Display lists are used to prepare the shape in outline and fill mode
        self._display_lists = []
        
        self.prepare()
        
    def __del__(self):
        # Remove any associated display lists
        for display_list in self._display_lists:
            glDeleteLists(display_list, 1)

    # points property
    def _get_points(self):
        return self._points
    
    def _set_points(self, value):
        self._points = value
        self.prepare()
    
    points = property(_get_points, _set_points)

    @property
    def mid_point(self):
        min_x = 1e+10
        max_x = -1e+10
        min_y = 1e+10
        max_y = -1e+10
        
        for point in self._points:
            if point[0] < min_x:
                min_x = point[0]
            if point[0] > max_x:
                max_x = point[0]
            
            if point[1] < min_y:
                min_y = point[1]
            if point[1] > max_y:
                max_y = point[1]
                
        return (float(min_x + max_x) / 2.0, float(min_y + max_y) / 2.0)

    # handle property
    def _get_handle(self):
        return self._handle
    
    def _set_handle(self, value):
        self._handle = value
        self.prepare()
    
    handle = property(_get_handle, _set_handle)
    
    def center_handle(self):
        """Center the sprites handle""" 
        mid = self.mid_point
        self.handle = (-mid[0],-mid[1])
    
    @property
    def smooth(self):
        return self._smooth
    
    @property
    def bounding_box(self):
        min_x = 1e+10
        max_x = -1e+10
        min_y = 1e+10
        max_y = -1e+10

        for point in self._point_list:
            if point[0] < min_x:
                min_x = point[0]
            if point[0] > max_x:
                max_x = point[0]

            if point[1] < min_y:
                min_y = point[1]
            if point[1] > max_y:
                max_y = point[1]

        return (min_x, min_y, max_x - min_x, max_y - min_y)

    @property
    def bounding_radius(self):
        mid = self.mid_point
        max_radius = -1e+10
        
        for point in self._points:
            point = (abs(mid[0] - point[0]), abs(mid[1] - point[1]))
            radius = sqrt(pow(point[0], 2) + pow(point[0], 2))
            if radius > max_radius:
                max_radius = radius
        
        return radius

    def get_abs_points(self):
        """Return the shapes points converted to their absolute positions"""
        abs_points = []
        
        x = self.x
        y = self.y
        s = self.scale
        h = (self._handle[0] * s, self._handle[1] * s)
        a = radians(-self.angle)
        for point in self._points:
            proj_point = ((point[0] * s) + h[0], (point[1] * s) + h[1])
            proj_angle = atan2(proj_point[0], proj_point[1]) + a
            proj_distance = sqrt(pow(proj_point[0], 2) + pow(proj_point[1], 2))
            abs_points.append((
                x + point[0] + (sin(proj_angle) * proj_distance),
                y + point[1] + (cos(proj_angle) * proj_distance)
                ))

        return abs_points

    def prepare(self):
        """Prepare the shape for drawing"""
        
        # Remove any existing display lists
        for display_list in self._display_lists:
            glDeleteLists(display_list, 1)
        self._display_lists = []
        
        # Prepare outline
        outline_display_list = glGenLists(1)
        glNewList(outline_display_list, GL_COMPILE)

        if len(self._points) > 1:
            if self._smooth:
                glEnable(GL_LINE_SMOOTH) 
            else:
                glDisable(GL_LINE_SMOOTH)

            glBegin(GL_LINE_LOOP)
            for point in self._points:
                glVertex2f(point[0] + self._handle[0], -(point[1] + self._handle[1]))

            glEnd()
            glEndList()

        elif len(self._points) == 1:
            if self._smooth:
                glEnable(GL_POINT_SMOOTH) 
            else:
                glDisable(GL_POINT_SMOOTH)
            
            glBegin(GL_POINTS)
            glVertex2f(
                self.points[0][0] + self._handle[0], 
                -(self.points[0][1] + self._handle[1])
                )
            glEnd()
            glEndList()
            
        self._display_lists.append(outline_display_list)

        # Prepare fill
        if len(self._points) > 2:
            fill_display_list = glGenLists(1)
            glNewList(fill_display_list, GL_COMPILE)

            glBegin(GL_POLYGON)
            for point in self._points:
                glVertex2f(
                    point[0] + self._handle[0], 
                    -(point[1] + self._handle[1])
                    )
            glEnd()
            glEndList()

            self._display_lists.append(fill_display_list)
    
    def draw(self):
        """Draw the shape to the screen"""
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        
        # Transform the sprite
        glTranslatef(self.x, Display().flip_y(self.y), 1.0)
        glRotatef(self.angle, 0.0, 0.0, -1.0)
        glScalef(self.scale, self.scale, 1.0)  

        # Draw fill
        if self.fill_color is not None and len(self._display_lists) == 2:
            glColor4ub(*self.fill_color)
            glCallList(self._display_lists[1])
            
        # Draw outline
        if self.outline_color is not None:
            glColor4ub(*self.outline_color)
            glCallList(self._display_lists[0])  


class Point(Shape):
    """
    A point.
    """

    def __init__(self, *args, **kw):
        super(Point, self).__init__([(0, 0)], *args, **kw)


class Line(Shape):
    """
    A line.
    """

    def __init__(self, length, *args, **kw):
        self._length = length
        super(Line, self).__init__([(0, 0), (length, 0)], *args, **kw)

    @property
    def length(self):
        return self._length


class Rect(Shape):
    """
    A rectangle.
    """

    def __init__(self, size, *args, **kw):
        self._size = size
        super(Rect, self).__init__([
            (0, 0), 
            (size[0], 0),
            (size[0], size[1]), 
            (0, size[1])
            ], *args, **kw)

    @property
    def size(self):
        return self._size

    def prepare(self):
        """Prepare the shape for drawing"""
        
        # Remove any existing display lists
        for display_list in self._display_lists:
            glDeleteLists(display_list, 1)
        self._display_lists = []

        # Prepare outline
        outline_display_list = glGenLists(1)
        glNewList(outline_display_list, GL_COMPILE)

        if self._smooth:
            glEnable(GL_LINE_SMOOTH) 
        else:
            glDisable(GL_LINE_SMOOTH)

        points = self._points
            
        glBegin(GL_LINE_LOOP)
        glVertex2f(points[0][0] + self._handle[0], -(points[0][1] + self._handle[1] + 1))
        glVertex2f(points[1][0] + self._handle[0] - 1, -(points[1][1] + self._handle[1] + 1))
        glVertex2f(points[2][0] + self._handle[0] - 1, -(points[2][1] + self._handle[1]))
        glVertex2f(points[3][0] + self._handle[0], -(points[3][1] + self._handle[1]))

        glEnd()
        glEndList()

        self._display_lists.append(outline_display_list)

        # Prepare fill
        fill_display_list = glGenLists(1)
        glNewList(fill_display_list, GL_COMPILE)

        glBegin(GL_TRIANGLE_FAN)
        glVertex2f(points[3][0] + self._handle[0], -(points[3][1] + self._handle[1]))
        glVertex2f(points[2][0] + self._handle[0], -(points[2][1] + self._handle[1]))
        glVertex2f(points[1][0] + self._handle[0], -(points[1][1] + self._handle[1]))
        glVertex2f(points[0][0] + self._handle[0], -(points[0][1] + self._handle[1]))
        glEnd()
        glEndList()

        self._display_lists.append(fill_display_list)


class Ellipse(Shape):
    """
    An ellipse.
    """
    
    def __init__(self, size, steps=36, *args, **kw):
        self._size = size
        
        half_size = (float(size[0]) / 2.0, float(size[1]) / 2.0)
        points = []
        for step in xrange(0, steps):
            angle = step * (2 * pi) / float(steps)
            pos = (half_size[0] * sin(angle), half_size[1] * cos(angle))
            points.append((half_size[0] + pos[0], half_size[1] + pos[1]))

        super(Ellipse, self).__init__(points, *args, **kw)
    
    @property
    def size(self):
        return self._size


class Polygon(Shape):
    """
    A polygon.
    """

    def save(self, f):
        """Save the polygon"""
        stream = True
        if isinstance(f, basestring):
            f = open(f, 'wb')
            stream = False
            
        f.write(struct.pack('i', len(self._points)))
        for point in self._points:
            f.write(struct.pack('f', point[0]))
            f.write(struct.pack('f', point[1]))
        
        if stream:
            f.close()
        
    @classmethod
    def load(cls, f, smooth=False):
        """Load a polygon from file"""
        stream = True
        if isinstance(f, basestring):
            f = open(f, 'rb')
            stream = False
            
        point_count = struct.unpack('i', f.read( 4 ) )[0]
        points = []
        for point_index in xrange(0, point_count):
            point = (
                struct.unpack('f', f.read(4))[0],
                struct.unpack('f', f.read(4))[0]
                )
            points.append(point)     

        if stream:
            f.close()
        
        return Polygon(points, smooth=smooth)
