"""
Basic UI elements.
"""

from hitt.collision import point_in_rect
from hitt.hid import HID
from hitt.shape import Rect
from hitt.utils import trigger_event


__all__ = (
    'UI',
    'UIWidget',
    'TextButton'
    )


class UI(object):
    """
    A class to handle events and drawing of a UI.
    """
    
    def __init__(self):
        self._widgets = {}
        self._blocked = {}
        
    def add(self, widget):
        """Add a widget to the UI"""
        self._widgets[widget._id] = widget

    def handle_events(self):
        """Handle UI events"""
        self._blocked = {}
        for widget in self._widgets.values():
            widget.handle_events()
    
    def block(self, event_type):
        """Mark an event type as being prevented until the next loop"""
        self._blocked[event_type] = True
    
    def blocked(self, event_type):
        return event_type in self._blocked
        
    def draw(self):
        """Draw the UI"""
        for widget in self._widgets.values():
            widget.draw()


class UIWidget(object):
    """
    A base class for all UI widgets.
    """
    
    _auto_inc = 0
    
    def __init__(self):
        
        # Assign the ID
        UIWidget._auto_inc += 1
        self._id = UIWidget._auto_inc
        
        self._states = {}
        
        self.x = 0
        self.y = 0
        self._size = [0, 0]
    
    @property
    def size(self):
        return self._size
    
    def state(self, state):
        """Return True if the widget is in the specified state"""
        return self._states.get(state, False)
    
    def handle_events(self):
        """Handle events for the widget"""
    
    def draw(self):
        """Draw the widget"""


class TextButton(UIWidget):
    """
    A simple text button class.
    """
    
    PADDING = 4
    
    def __init__(self, label, font, 
        bg_color=(0, 0, 0, 96), 
        fg_color=(255, 255, 255, 255)):
        
        super(TextButton, self).__init__()
        
        self._states = {
            'over': False,
            'down': False
        }
        
        self._label = label
        self._font = font
        self.bg_color = bg_color
        self.fg_color = fg_color
        self._bg = None
        
        self._prepare()
    
    def _prepare(self):
        # Recalculate the size of the button
        pad = TextButton.PADDING
        self._size = [dim + pad for dim in self._font.measure(self._label)]
        self._bg = Rect(self._size)
        
    # Label property
    def _get_label(self):
        return self._label
        
    def _set_label(self, label):
        self._label = label
        self._prepare()
    
    label = property(_get_label, _set_label)
    
    # Font property
    def _get_font(self):
        return self._font

    def _set_font(self, label):
        self._font = font
        self._prepare()

    font = property(_get_font, _set_font)
    
    @trigger_event
    def click(self):
        """Trigger the click event"""
    
    def handle_events(self):
        """Handle events for the button"""
        
        if point_in_rect(HID().mouse.pos, (self.x, self.y), self._size):
            self._states['over'] = True
            if HID().mouse.button_down(0):
                self._states['down'] = True
            else:
                if self._states['down']:
                    self.click()
                self._states['down'] = False
        else:
            self._states['over'] = False
            self._states['down'] = False
    
    def draw(self):
        """Draw the button"""
        
        # Draw the background
        self._bg.x = self.x
        self._bg.y = self.y
        self._bg.outline_color = None
        self._bg.fill_color = self.bg_color
        self._bg.draw()
        
        # Draw the label
        inner_pad = TextButton.PADDING / 2
        self._font.draw(self._label, 
            (self.x + inner_pad, self.y + inner_pad),
            color=self.fg_color,
            cache=True
            )
        
        