"""
Support for rendering a grid.
"""

import Image as PILImage, ImageDraw as PILImageDraw, ImageFont as PILImageFont
from OpenGL.GL import *

from hitt.display import Display
from hitt.utils import img2tex, nearest_pow2


class Grid(object):
    """
    The Grid class allows you to render a 2D graph.
    """
    
    def __init__(self, size=(32, 32)):
        self._size = size
        self.color = (255, 255, 255, 255)
        
        self.x = 0
        self.y = 0
        self._texture = None
        self._display_list = None
        
        self.prepare()
        
    def __del__(self):
        # Remove any associated texture
        if self._texture is not None:
            glDeleteTextures([self._texture])

        # Remove any associated display list
        if self._display_list is not None:
            glDeleteLists(self._display_list, 1)
        
    @property
    def size(self):
        return self._size
    
    def prepare(self):
        """Prepare the grid for rendering"""
        
        # Remove any associated texture
        if self._texture is not None:
            glDeleteTextures([self._texture])

        # Remove any associated display list
        if self._display_list is not None:
            glDeleteLists(self._display_list, 1)
    
        # Generate the grid image
        w = nearest_pow2(self.size[0])
        h = nearest_pow2(self.size[1])
        image = PILImage.new('L', (w, h))
        draw = PILImageDraw.Draw(image)
        draw.line((0, 0, w, 0), fill=255)
        draw.line((0, 0, 0, h), fill=255)
        del draw
        
        # Convert the image to a texture
        self._texture = img2tex(image, clamp=False)
        
        # Create a display list to store the grid drawing operations in
        if self._display_list is not None:
            glDeleteLists(self._display_list, 1)
        self._display_list = glGenLists(1)
        glNewList(self._display_list, GL_COMPILE)
        
        # Pre-render the sprite
        glPolygonMode(GL_FRONT, GL_FILL)      
        glEnable(GL_TEXTURE_2D)  
        
        # Assign the texture
        glBindTexture(GL_TEXTURE_2D, self._texture)
       
        w, h = Display().size
        w = (self._size[0] * 2) + self._size[0] * int(float(w) / float(self._size[0]))
        h = (self._size[1] * 2) + self._size[1] * int(float(h) / float(self._size[1]))
        x_repeat = float(w) / float(self._size[0])
        y_repeat = float(h) / float(self._size[1])
       
        # Draw the quad
        glBegin(GL_TRIANGLE_FAN)
        
        glTexCoord2f(0, y_repeat)
        glVertex2f(0, -h)
        
        glTexCoord2f(x_repeat, y_repeat)
        glVertex2f(w, -h)
        
        glTexCoord2f(x_repeat, 0)
        glVertex2f(w, 0)
        
        glTexCoord2f(0, 0)
        glVertex2f(0, 0)
        
        glEnd()
        glDisable(GL_TEXTURE_2D)
        glEndList()

    def draw(self):
        """Draw the grid on screen"""
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        # Wrap values
        x = self.x % self._size[0]
        y = self.y % self._size[1]

        # Transform the grid
        glTranslatef(x - self._size[0], Display().flip_y(y - self._size[1]), 1.0)

        # Render the grid
        glColor4ub(*self.color)
        glCallList(self._display_list)        
    