"""
A set of utility classes, decorators and functions.
"""

import Image as PILImage
import math
from OpenGL.GL import *
from pydispatch import dispatcher
from pygame import display as _pygame_display
import struct


__all__ = (
    # Classes
    'MixinMeta',
    'SingletonMeta',
    
    # Decorators
    'trigger_event',
    
    # Functions
    'bin2str',
    'bin2img',
    'img2bin',
    'str2bin',
    'img2tex',
    'nearest_pow2',
    'split_pow2',
    )


# Classes

class MixinMeta(type):
    """
    The MixinMeta class allows you to specify a list of classes to 'mix in' 
    with the classe using the the __mixins__ attribute.
    """
    
    def __new__(cls, name, bases, attrs):

		mixin_attrs = {}
		if '__mixins__' in attrs:
		    for mixin_cls in attrs['__mixins__']:
		        for attr_name in mixin_cls.__dict__:
		            if not attr_name.startswith('__'):
		                mixin_attrs[attr_name] = mixin_cls.__dict__[attr_name]
		
		attrs.update(mixin_attrs)
		
		new_cls = super(MixinMeta, cls).__new__(cls, name, bases, attrs)
		
		return new_cls
		
		
class SingletonMeta(type):
    """
    A singleton meta class to simplify the process of creating a singleton 
    class.
    """
    
    def __init__(cls, name, bases, dict):
        super(SingletonMeta, cls).__init__(name, bases, dict)
        cls._instance = None 

    def __call__(cls,*args,**kw):
        if cls._instance is None:
            cls._instance = super(SingletonMeta, cls).__call__(*args, **kw)
        return cls._instance


# Decorators

def trigger_event(func):
    """A decorator to that triggers events when a class method is called"""
    def _trigger_event(self, *args, **kwargs):
        dispatcher.send(signal='on_' + func.__name__, sender=self)
        func(self, *args, **kwargs)
        dispatcher.send(signal='after_' + func.__name__, sender=self)
        
    return _trigger_event


# Functions    

def bin2img(data):
    """Return an image loaded from binary data"""
    
    # Image mode
    image_mode = bin2str(data)
    data = data[len(image_mode) + struct.calcsize('I'):]
    
    # Image size
    image_size = struct.unpack('HH', data[:struct.calcsize('HH')])
    data = data[struct.calcsize('HH'):]
    
    # Image data
    image_data = data[:image_size[0] * image_size[1] * len(image_mode)]
    return PILImage.fromstring(image_mode, image_size, image_data)
    
def bin2str(data):
    """Return a string loaded from binary data"""
    offset = struct.calcsize('I')
    length = struct.unpack('I', data[0:offset])[0]
    return data[offset:offset + length]
    
def img2bin(img):
    """Return an image saved in binary format"""
    data = str2bin(img.mode)
    data += struct.pack('HH', *img.size)
    data += img.tostring()
    return data

def str2bin(chars):
    """Return a string saved in binary format"""
    return struct.pack('I', len(chars)) + chars.encode('ascii')
    
def img2tex(image, clamp=True, smooth=False):
    """
    Convert a PIL image to an OpenGL texture.
    """
    
    # Ensure the image is a suitable texture size
    max_size = glGetInteger(GL_MAX_TEXTURE_SIZE)
    w = nearest_pow2(min(max_size, image.size[0]))
    h = nearest_pow2(min(max_size, image.size[1]))
    
    # If the image is not a suitable size resize the image to fit
    if (w, h) != image.size:
        image = image.resize((w, h), PILImage.BICUBIC)
    
    # Generate the OpenGL texture
    texture = glGenTextures(1)
    
    # Set the texture's flags
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1)
    glBindTexture(GL_TEXTURE_2D, texture)
    if clamp:
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
        
    glTexParameteri(GL_TEXTURE_2D, 
        GL_TEXTURE_MAG_FILTER, 
        GL_LINEAR if smooth else GL_NEAREST)
    glTexParameteri(GL_TEXTURE_2D, 
        GL_TEXTURE_MIN_FILTER, 
        GL_LINEAR if smooth else GL_NEAREST)
    
    # Prepare and apply the image data 
    pil_mode = 'RGBA'
    gl_mode = GL_RGBA
    if image.mode == 'L':
        pil_mode = 'L'
        gl_mode = GL_ALPHA
    elif image.mode == "RGB":
        pil_mode = 'RGBX'
        
    glTexImage2D(GL_TEXTURE_2D, 
        0, gl_mode, 
        w, h, 
        0, gl_mode, 
        GL_UNSIGNED_BYTE, image.tostring('raw', pil_mode, 0, -1))
                    
    return texture

def nearest_pow2(value):
    """Find the nearest power of two (linearly) to a given number"""
    return int(math.pow(2, math.ceil(math.log(value, 2))))

def split_pow2(value):
    """
    Split a value down into the optimum list of power of two values, (e.g 
    power of 2 in size e.g 144 -> [128,16]).
    """
    divs = []
    div = glGetInteger(GL_MAX_TEXTURE_SIZE)
    
    while div >= 1:
        # Detemine the number of divisions at the current size
        div_count = int(float(value) / div)
        
        for i in xrange(0, div_count):
            divs.append(div)
        
        if (value % div) > 0:
            value -= div_count * div
            div = div / 2
        else:
            div = 0
        
    return divs



    